from microprediction.writer import MicroWriter
from microprediction.conventions import api_url
import time
from pprint import pprint
import datetime
import threading
import os


class MicroPoll(MicroWriter):

    # --------------------------------------------
    #  You may wish to override these methods
    # --------------------------------------------

    def logger(self, data):
        """ Called after each attempt to poll data and send it """
        self.recent.append(data)
        if len(self.recent) > 100:
            self.recent = self.recent[-100:]
        if self.verbose:
            pprint(data)
            print(' ', flush=True)

    def downtime(self):
        """ Also called after each attempt to send data """
        # By default it does a little MUID mining, or a little more if the balance is sinking
        self.maybe_bolster_balance_by_mining()

    def determine_next_value(self, source_value):
        """ Receives source data and decides what to send, if anything """
        # This should return a float or None
        return float(source_value) if source_value is not None else None

    # --------------------------------------------
    #  Probably don't want to override the rest
    # --------------------------------------------

    def __init__(self, name, func, interval, write_key="invalid_key", base_url=None, verbose=True, func_args=None,
                 mine=False, **kwargs):
        """  Create a stream by polling every 20 minutes, say
            param: name  str   stream name ending in .json
            get_iex_realtime_price:        function    returns float  (data feed function)
            interval:    int     minutes between polls
            func_args    dict    optional dict of arguments to be passed to get_iex_realtime_price
            mine:        bool    Supply mine=True if you want to automatically mine MUIDs to help keep the stream alive
        """
        assert self.is_valid_name(name), 'name not valid'
        super().__init__(base_url=base_url or api_url(), write_key=write_key, verbose=verbose)
        self.name = name
        self.interval = interval
        self.func = func
        self.func_args = func_args
        self.recent = list()
        self.mining_time = 0
        self.mining_success = 0
        self.test_value = self.call_func()  # Fail fast
        self.mine = mine
        print('Created poller. Example value ' + str(self.test_value))

    def __repr__(self):
        self_data = {'get_iex_realtime_price': str(self.func), 'interval': self.interval,
                     'mining_time': self.mining_time}
        if self.func_args is not None:
            self_data.update(self.func_args)
        return self_data

    def call_func(self):
        if isinstance(self.func_args, list):
            return self.func(*self.func_args)
        elif isinstance(self.func_args, dict):
            return self.func(**self.func_args)
        else:
            return self.func()

    def task(self):
        """ Scheduled task that runs every interval minutes """
        start_time = time.time()
        data = {'start_time': str(datetime.datetime.now())}
        try:
            source_value = self.call_func()
        except Exception as e:
            source_value = None
        next_value = self.determine_next_value(source_value)
        data.update({'source_value': source_value,
                     'next_value': next_value,
                     'elapsed after polling': time.time() - start_time})
        res = self.touch(self.name) if next_value is None else self.set(name=self.name, value=next_value)
        data.update({'value': next_value, "res": res, 'elapsed after sending': time.time() - start_time})
        self.logger(data=data)
        self.downtime()
        data.update({'elapsed after downtime': time.time() - start_time})

    def maybe_bolster_balance_by_mining(self):
        """ Mine just a little to avoid stream dying due to bankruptcy """
        if self.mine:
            balance = self.get_balance()
            if balance < 0:
                muid_time = time.time()
                key = self.bolster_balance_by_mining(seconds=max(5, int(abs(balance) / 10)))
                mining_time = time.time() - muid_time
                self.mining_time += mining_time
                if key:
                    print('************************')
                    print('     FOUND MUID !!!     ')
                    print('************************', flush=True)
                    self.mining_success += 1
                else:
                    print('Did not find MUID this time', flush=True)

    def run(self, timeout=60 * 60 * 24 * 365, **ignore):
        """
            Simple single task scheduler that skips if task takes too long
              :param timeout second to run for
        """
        run_start_time = time.time()
        run_end_time = run_start_time + timeout
        while time.time() < run_end_time:
            st = int(round(time.time()))
            self.task()
            et = int(round(time.time()))
            sleep_time = (st - et) % (60 * self.interval)
            time.sleep(sleep_time)


class ChangePoll(MicroPoll):
    #    Illustrates predicting the change in a quantity ... when feed is not entirely reliable
    #    If we don't get good data from the feed or it appears to be stale, we don't publish the change

    COLD = 0
    WARM = 1

    def __init__(self, name, func, interval, write_key, func_args=None):
        super().__init__(name=name, func=func, interval=interval, func_args=func_args, write_key=write_key)
        self.prev_value = None
        self.feed_state = ChangePoll.COLD
        self.current_value = None

    def alert(self, message):
        print(message, flush=True)

    def determine_next_value(self, source_value):
        """ Returns a change if feed is warm, else None """

        if self.feed_state == ChangePoll.WARM:
            # A warm state means that previous speed exists and is not stale
            assert self.prev_value is not None
            self.current_value = source_value
            if self.current_value is None:
                self.feed_state = ChangePoll.COLD
                self.alert(message='Something amiss with feed')
                return None
            else:
                self.current_value = float(source_value)
                value_change = float(self.current_value) - float(self.prev_value)
                if abs(value_change) < 1e-5:
                    self.feed_state = ChangePoll.COLD  # Feed is stale, don't judge
                    self.logger(
                        {'type': 'feed_status', 'message': "****  Feed unchanged at " + str(datetime.datetime.now())})
                else:
                    self.prev_value = self.current_value
                    return value_change
        elif self.feed_state == ChangePoll.COLD:
            # Wait until feed is back up and speeds start changing
            self.prev_prev = self.prev_value if self.prev_value else None
            self.prev_value = source_value
            if (self.prev_value is not None) and (self.prev_prev is not None) and abs(
                    float(self.prev_prev) - float(self.prev_value)) > 1e-5:
                self.feed_state = ChangePoll.WARM
                self.logger({'type': 'feed_status', 'message': '**** Feed resumed at ' + str(datetime.datetime.now())})
            return None


class MultiPoll(MicroPoll):

    # Create multiple streams updated at the same time

    # -------------------------------------------
    #  You may wish to override this method if the
    #  data retrieval function get_iex_realtime_price() you provide
    #  does not return [ float ] or [ str ] that is
    #  convertable to float
    # --------------------------------------------

    def determine_next_values(self, source_values):
        """ Should receive raw source data and decides what to send, if anything

            :param source_values   [ float ] or whatever type is returned by self.get_iex_realtime_price. Simplest would be [ float ]
            :returns  [ float ]  or None
        """

        if source_values is None or (any(source_value is None for source_value in source_values)):
            return None
        else:
            return [float(sv) for sv in source_values]

    # --------------------------------------------
    #  Maybe don't override the rest
    # --------------------------------------------

    def __init__(self, names, func, interval, write_key="invalid_key", base_url=None, verbose=True, func_args=None,
                 with_copulas=False, secondary_func=None, secondary_func_args=None, **kwargs):
        """  Create a stream by polling every 20 minutes, say
                param: names  [ str ]    stream name ending in .json
                get_iex_realtime_price:        function    returns raw data from some live source, ideally [ float ] but you can override determine_next_values method
                interval:    int         minutes between polls
                func_args    dict        optional dict of arguments to be passed to get_iex_realtime_price

            """
        assert all([self.is_valid_name(name) for name in names]), 'name not valid'
        super().__init__(base_url=base_url or api_url(), write_key=write_key, verbose=verbose, func=func,
                         func_args=func_args, interval=interval, name='never_used.json')
        self.names = names
        self.with_copulas = with_copulas
        self.secondary_func = secondary_func
        self.secondary_func_args = secondary_func_args

    def call_secondary_func(self, next_values):
        if isinstance(self.secondary_func_args, list):
            return self.secondary_func(next_values, *self.secondary_func_args)
        elif isinstance(self.secondary_func_args, dict):
            return self.secondary_func(next_values, **self.secondary_func_args)
        else:
            return self.secondary_func(next_values)

    def task(self):
        """ Scheduled task that runs every minute """
        start_time = time.time()
        data = {'start_time': str(datetime.datetime.now())}
        try:
            source_values = self.call_func()
        except Exception as e:
            source_values = None
        next_values = self.determine_next_values(source_values)
        data.update({'source_values': source_values,
                     'next_values': next_values,
                     'elapsed after polling': time.time() - start_time})

        # Update the values with secondary get_iex_realtime_price
        if (self.secondary_func is not None) and (next_values is not None):
            send_values = self.apply_secondary_func(next_values)
        else:
            send_values = next_values

        # Send the data
        if next_values is None:
            res = {'operation': 'touch', 'exec': [self.touch(name=name) for name in self.names]}
        elif self.with_copulas:
            res = self.cset(names=self.names, values=send_values)
        else:
            res = [self.set(name=name, value=value) for name, value in zip(self.names, send_values)]

        data.update({'values': next_values, "res": res, 'elapsed after sending': time.time() - start_time})
        self.logger(data=data)
        self.downtime()
        data.update({'elapsed after downtime': time.time() - start_time})


def default_change_criterion(old_values, new_values):
    [abs(pp - pv) > 1e-6 for pp, pv in zip(old_values, new_values)]


class MultiChangePoll(MultiPoll):
    COLD = 0
    WARM = 1

    def __init__(self, names, func, interval, write_key="invalid_key", base_url=None, verbose=True, func_args=None,
                 with_copulas=False, change_func=None, change_func_args=None, **kwargs):
        """  Create multiple streams by polling every 20 minutes, say
                param:
                names        [ str ]    stream name ending in .json
                get_iex_realtime_price:        function    returns data from some live source, ideally [ float ] but you can override determine_next_values method
                interval:    int         minutes between polls
                func_args    dict        optional dict of arguments to be passed to get_iex_realtime_price
                with_copulas bool        Whether to create derived copula streams, or just individual unrelated streams
                change_func              Function acting directly on a list of value changes
                change_func_args         Additional argument to change function
            """
        super().__init__(base_url=base_url or api_url(), write_key=write_key, verbose=verbose, func=func,
                         func_args=func_args, interval=interval, names=names, with_copulas=with_copulas, **kwargs)
        self.prev_values = None
        self.feed_state = MultiChangePoll.COLD
        self.current_values = None
        self.change_func = change_func
        self.change_func_args = change_func_args

    def call_change_func(self, value_changes: [float]) -> [float]:
        if isinstance(self.change_func_args, list):
            return self.change_func(value_changes, *self.change_func_args)
        elif isinstance(self.change_func_args, dict):
            return self.change_func(value_changes, **self.change_func_args)
        else:
            return self.change_func(value_changes)

    def alert(self, message):
        print(message, flush=True)

    def determine_next_values(self, source_values):
        if self.feed_state == MultiChangePoll.WARM:
            # A warm state means that previous speed exists and is not stale
            assert self.prev_values is not None
            self.current_values = source_values
            if self.current_values is None:
                self.feed_state = MultiChangePoll.COLD
                self.alert(message='Something amiss with feed')
                return None
            else:
                self.current_values = [float(source_value) for source_value in source_values]
                value_changes = [float(current_value) - float(prev_value) for current_value, prev_value in
                                 zip(self.current_values, self.prev_values)]
                material_changes = [abs(vc) > 1e-6 for vc in value_changes]
                if not any(material_changes):
                    self.feed_state = MultiChangePoll.COLD  # Feed is stale, don't judge
                    self.logger(
                        {'type': 'feed_status', 'message': "****  Feed unchanged at " + str(datetime.datetime.now())})
                else:
                    self.prev_values = self.current_values
                    if self.change_func is not None:
                        altered_changes = self.call_change_func(value_changes=value_changes)
                    else:
                        altered_changes = [c for c in value_changes]
                    return altered_changes

        elif self.feed_state == MultiChangePoll.COLD:
            # Wait until feed is back up and values start changing
            self.prev_prev = self.prev_values if self.prev_values else None
            self.prev_values = source_values
            if (self.prev_values is not None) and (self.prev_prev is not None):
                material_changes = [abs(pp - pv) > 1e-6 for pp, pv in zip(self.prev_prev, self.prev_values)]
                if any(material_changes):
                    self.feed_state = MultiChangePoll.WARM
                    self.logger(
                        {'type': 'feed_status', 'message': '**** Feed resumed at ' + str(datetime.datetime.now())})
            return None
        else:
            raise Exception('Brain failure')



# IEX stock prices
# You can get your api_key at https://iexcloud.io/
# See https://iexcloud.io/docs/api/#quote


def iex_latest_prices(tickers, api_key:str)->[float]:
    """
        Grab multiple stock prices
    """
    symbols = ','.join(tickers)
    paid_url = 'https://cloud.iexapis.com/v1/stock/market/batch?types=price&symbols='+symbols+'&token=' + api_key
    print(paid_url)
    data = getjson(paid_url)
    mids = [ data[ticker.upper()]['price'] for ticker in tickers]
    return mids


if __name__ == '__main__':
    from pprint import pprint
    from getjson import getjson
    from microprediction.config_private import IEX_KEY
    tickers = ['aapl','googl']
    data = iex_latest_prices(tickers=tickers, api_key=IEX_KEY)
    pprint(data)


from microprediction.univariate.distmachine import DistMachine
from microconventions.stats_conventions import StatsConventions
from microprediction.univariate.runningmoments import RunningVariance


# Maintain running estimate of normal distribution

class NormalDist(RunningVariance, DistMachine):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def inv_cdf(self, p: float) -> float:
        return self.mean + StatsConventions.norminv(p) * self.std()




import numpy as np
import os
import json
import math
from itertools import combinations_with_replacement
from microprediction.live.iex import iex_latest_prices
from microprediction import MicroReader
from scipy.stats import kurtosis

# A few conventions used by FAANG streams


FAANG_TICKERS = ['fb', 'aapl', 'amzn', 'nflx', 'googl']
FAANG_NAMES = ['faang_' + str(k) + '.json' for k,ticker in enumerate(FAANG_TICKERS)] # Stream names
N_FAANG = len(FAANG_TICKERS)
FAANG_COMBINATIONS = list(combinations_with_replacement(range(N_FAANG), r=5))


def iex_scaled_log_faang(api_key:str):
    """
       Retrieve live prices for FAANG and transform
       api_key: An IEX api key
    """
    sp = iex_latest_prices(tickers=FAANG_TICKERS, api_key=api_key)
    return [1000 * math.log(v) for v in sp]


def portfolio_from_combination(c):
    """
       Converts a combination like 1,1,1,2,3 into weighted portfolio
    """
    n = len(c)
    w = 0.5*np.ones(n)/n
    for ndx in c:
        w[ndx]+=0.5/n
    return w


def readable_portfolio_from_combination(c,tickers):
    w = portfolio_from_combination(c)
    return 'Portfolio comprising '+', '.join( [ str(round(wi*100,0))+'% '+ticker for wi,ticker in zip(w,tickers)])


def gnaff_name_from_combination(c):
    return 'gnaaf_'+''.join([ str(cj) for cj in c]) + '.json'


GNAAF_NAMES = [gnaff_name_from_combination(c) for c in FAANG_COMBINATIONS]
GNAAF_WEIGHTS = [ portfolio_from_combination(c) for c in FAANG_COMBINATIONS]
GNAAF_LISTING = dict([(readable_portfolio_from_combination(c, tickers=FAANG_TICKERS), 'https://www.microprediction.org/stream_dashboard.html?stream=' + gnaff_name_from_combination(c).replace('.json', '')) for c in FAANG_COMBINATIONS])

NAAFG_METHODS = [('info', 1),  # The second entry is an exponent
                 ('info', 2),
                 ('info', 4),
                 ('kurtosis', 1),
                 ('kurtosis', 2),
                 ('kurtosis', 4),
                 ('semi', 1),
                 ('semi', 2),
                 ('semi', 4)]
NAAFG_NAMES = ['naafg_' + str(k) + '.json' for k, _ in enumerate(NAAFG_METHODS)] # Community portfolios




def create_listing_file():
    """
       Updates the stream listing
    """
    from microprediction.whereami import ROOT
    fn = os.path.join(ROOT, 'stream_examples_faang', 'stream_list.json')
    with open(fn,'wt') as fh:
        json.dump(GNAAF_LISTING,fh)


def scaled_portfolio_return(changes, w:[float]):
    """
       Compute scaled portfolio log changes from scaled asset log changes
    """
    w = np.array(w)/np.sum(w)   # Ensure weights normalized
    f = np.exp( np.array(changes)/1000.)  # Multiplicative increase in wealth
    w_post = float( np.inner( w, f) )    # Posterior unit portfolio wealth
    portfolio_change = 1000*math.log( w_post )
    return portfolio_change


def get_gnaff_prediction_metrics(write_key):
    """
       Get metrics for current predictions for GNAAF streams
       (Only the stream owner can do this)
    """
    the_means = list()
    the_stds = list()
    the_semis = list()
    the_kurtosis = list()
    reader = MicroReader()
    for name in GNAAF_NAMES:
        try:
            all_predictions = reader.get_predictions(name=name, write_key=write_key, delay=reader.DELAYS[2], strip=True, consolidate=True)
            trimmed_values = [v if abs(v) < 100. else 100 * v / abs(v) for v in all_predictions]
            n_chop = int(math.ceil(len(trimmed_values)) / 50)
            chopped = trimmed_values[n_chop:-n_chop]
            neg_chopped = [ 0 if (ch>=0) else ch*ch for ch in chopped ]
            the_means.append(np.mean(chopped))
            the_stds.append(np.std(chopped))
            the_semis.append(math.sqrt(np.mean(neg_chopped)))
            the_kurtosis.append(kurtosis(chopped))
        except Exception:
            print('Something wrong with get_gnaff_prediction_metrics')
            the_means.append(0)
            the_stds.append(1)
            the_kurtosis.append(-0.01)
            the_semis.append(1)
    assert len(the_means)+len(the_stds)==2*len(GNAAF_NAMES)
    return the_means, the_stds, the_semis, the_kurtosis


def naafg_community_portfolios(write_key)->[[float]]:
    """
        A collection of community constructed portfolios of FAANG stocks
           write_key: the private key for LEGLESS_OCELOT, creator of the GNAAF portfolio return streams
    """
    the_means, the_stds, the_semis, the_kurtosis = get_gnaff_prediction_metrics(write_key=write_key)
    scaled_infos = [ 5*mn/st for mn, st in zip(the_means, the_stds)]
    scaled_semi_infos = [ 5*mn/ss for mn,ss in zip(the_means, the_semis)]
    scaled_kurtosis = [ 0.7*kt for kt in the_kurtosis]

    the_portfolios = list()
    metrics = {'info':scaled_infos,'semi':scaled_semi_infos,
               'kurtosis':scaled_kurtosis}
    for metric_name,expon in NAAFG_METHODS:
        the_weighting = [ math.exp(expon*mtc) for mtc in metrics[metric_name] ]
        the_unnormalized_portfolio = np.array([ 0 for _ in FAANG_NAMES ])
        for weight, w in zip( the_weighting, GNAAF_WEIGHTS):
            the_unnormalized_portfolio = the_unnormalized_portfolio + weight*np.array(w)
        the_portfolio = list( the_unnormalized_portfolio/sum(the_unnormalized_portfolio) )
        the_portfolios.append(the_portfolio)
    return the_portfolios


if __name__=='__main__':
    create_listing_file()
    from microprediction.config_private import LEGLESS_OCELOT
    ports = naafg_community_portfolios(write_key=LEGLESS_OCELOT)
    from pprint import pprint
    pprint(ports)




from collections import OrderedDict
from microprediction.crawler import MicroCrawler
from microprediction.univariate.arrivals import approx_dt
from microprediction.samplers import normal_sample_projected, project_on_lagged_lattice, exponential_bootstrap_projected, center_and_scale_values_list
from microprediction.univariate.processes import k_std
from microprediction.samplers import fox_sample
import math
import numpy as np
from pprint import pprint

from microconventions.stats_conventions import evenly_spaced_percentiles, nudged

# Video tutorials are available at https://www.microprediction.com/python-1 to help you
# get started. There's a video explanation of FitCrawler, SequentialCrawler and friends
# at https://www.microprediction.com/fitcrawler

MAX_SKATER_K = 30


def split_k(approx_k:float):
    """ Represent a float 'k' as weighted combination of two integer ks  """
    approx_k = min(MAX_SKATER_K-0.1, approx_k)
    l = math.floor(approx_k)
    u = math.ceil(approx_k)
    r = approx_k - l
    return (l, 1-r ), (u, r)


class StreamSkater(MicroCrawler):

    # Crawler utilizing the timemachines package

    def __init__(self, f, n_warm:int=400, use_mean=True, use_std=True, decay=0.01, **kwargs):
        """
            f         A "skater" https://github.com/microprediction/timemachines
            n_warm    Number of historical data points to use the first time to warm up the skater
            use_mean  If True, the mean of samples matches the skater estimate prior to any lattice projection
            use_std   If True, we use the model standard deviation.
                        (Be aware that some models from the timemachines package give dubious std error estimates)

        See https://github.com/microprediction/timemachines for explanation of "skaters"
        A skater is a sequential online k-step ahead forecasting function.
        The timemachines package contains dozens of skaters, any one of which might be used here.
        """
        super().__init__(**kwargs)
        self.f = f
        self.n_warm = n_warm
        self.use_mean = use_mean
        self.use_std = use_std
        self.decay = decay
        self.stream_state = OrderedDict()


    #################################################
    #                                               #
    #   You'll likely want to override this ...     #
    #                                               #
    #################################################


    def sample_using_point_estimate(self, x:float, x_std:float, k, name, delay, lagged_values, lagged_times):
        """ Create 225 samples guided by a point estimate, and a standard deviation

              x      - point estimate returned by a function in the timemachines package (i.e. a "skater")
              x_std  - standard deviation of point estimate
              returns [float]

            By default this rescales recency weighted points and then projects onto a lattice of values that is also implied by the history
        """
        new_mean = x if self.use_mean else None
        new_std = x_std if self.use_std else None
        return exponential_bootstrap_projected(lagged_values=lagged_values, decay=self.decay,
                                       num=self.num_predictions, new_mean=new_mean, new_std=new_std)


    #####################################################
    #                                                   #
    #   You'll likely NOT need to override this ...     #
    #       (just skater maintenance)                   #
    #####################################################

    def sample(self, lagged_values, lagged_times=None, name=None, delay=None, **ignored):
        """ Use skater to move and scale """

        if name not in self.stream_state:
            self.stream_state[name] = {'skater_state':{},
                                       'x':None,
                                       'x_std':None,
                                       'dt':None,
                                       't':None,
                                       'lookup':None} # Map from delay

        state = self.stream_state[name]

        if state['dt'] is None:
            # Initialize lookups from delay to steps ahead
            state['dt'] = approx_dt(lagged_times)
            state['lookup'] = dict([(dly, split_k( max(1,0.1+dly / (0.01 + state['dt']))-1)) for dly in self.DELAYS])
            max_k = max( [ spl[1][0] for k, spl in state['lookup'].items() ]) + 1
            state['k'] = max_k  # max k

        # Determine which observations are yet to be processed by the skater
        if state['t'] is None:
            ys = reversed( lagged_values[:self.n_warm] )
            ts = reversed( lagged_times[:self.n_warm] )
        else:
            all_t = reversed(lagged_times)
            all_y = reversed(lagged_values)
            yt = [ (y_,t_) for y_,t_ in zip(all_y,all_t) if t_>state['t']+1e-6 ]
            ys = [ yt_[0] for yt_ in yt ]
            ts = [ yt_[1] for yt_ in yt ]

        # Run the skater
        for y_,t_ in zip(ys,ts):
            state['x'], state['x_std'], state['skater_state'] = self.f(y=y_,s=state['skater_state'],k=state['k'],a=None,t=t_,e=None)
            state['t'] = t_

        # Interpolate point estimate and std errors
        (low_k,low_k_weight), (high_k, high_k_weight) = state['lookup'][delay]
        try:
            x_interp = low_k_weight*state['x'][low_k] + high_k_weight*state['x'][high_k]
        except IndexError:
            print('!!!!!!!!!!!!!!!!!!!!!!!!!')
            print('Stream Skater interp failure')
            flr = {'low_k':low_k,'high_k':high_k,'low_k_weight':low_k_weight,'high_k_weight':high_k_weight,
                   'state[x]':state['x'],' len(state[x])':len(state['x']),'state':state}
            pprint(flr)
            x_interp = state['x'][low_k]

        try:
            x_std_interp = low_k_weight*state['x_std'][low_k] + high_k_weight*state['x_std'][high_k]
        except IndexError:
            x_std_interp = state['x_std'][low_k]

        # Save stream state for next invocation
        self.stream_state[name] = state

        # Create a hacky estimate of standard error, if necessary
        if not self.use_std:
            x_std_interp = k_std(lagged_values, k=high_k)

        return self.sample_using_point_estimate(x=x_interp, x_std=x_std_interp, k=high_k, name=name, delay=delay, lagged_values=lagged_values, lagged_times=lagged_times)



#################################################
#                                               #
#   And now for examples ...                  #
#                                               #
#################################################


class SkatingFox(StreamSkater):

    # Shows how to swap out the default method of creating 225 guesses from x, x_std skater predictions

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def sample_using_point_estimate(self, x:float, x_std:float, k:int, name:str, delay, lagged_values:[float], lagged_times):
        """ Create 225 samples guided by a point estimate, and a standard deviation
            Provided as an example of sub-classing StreamSkater, and modifying the sample generation.

                k  - Number of lags assumed to correspond to the delay

        """
        fox_samples = fox_sample(lagged_values=lagged_values, lagged_times=lagged_times, delay=delay,
                                 num=self.num_predictions, name=name)
        shifted_fox_samples = [ s + x - np.mean(fox_samples) for s in fox_samples ]
        r = x_std/np.std(fox_samples) if self.use_std else 1.0
        scaled_shifted_fox_samples = [ s + r*(s-x) for s in shifted_fox_samples ]
        return nudged(project_on_lagged_lattice(values=scaled_shifted_fox_samples, lagged_values=lagged_values))


class ChoosySkatingFox(SkatingFox):

    # Illustrates how to direct a StreamSkater to particular streams

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def exclude_stream(self, name=None, **ignore):
        # Example of influencing the navigation
        return '~' in name or 'emoji' in name

    def include_delay(self, delay=None, name=None, **ignore):
        # Example of influencing choice of horizons to predict
        return delay < 10*60


class RegularFaangStreamSkater(StreamSkater):

    # In particular ... this one hits regular streams only for the longest two horizons

    def __init__(self,**kwargs):
        super().__init__(**kwargs)

    def include_stream(self, name=None, **ignore):
        return ('~' not in name) and (('faang' in name) or ('gnaaf' in name))

    def include_delay(self, delay=None, name=None, **ignore):
        # Example of influencing choice of horizons to predict
        # We skip the shortest horizons
        return delay > self.DELAYS[1]



import os
from sys import platform
from getjson import getjson


# Retrieve credentials for database where state is stored

try:
    from microprediction.set_env_private import NOTHING_MUCH
except ImportError:
    # Environment variables need to be set some other way, for example by retrieving github secrets
    pass

micro_config_url = os.getenv('MICRO_TEST_CONFIG_URL')
if micro_config_url is None:
    raise Exception('Cannot get environment variable MICRO_TEST_CONFIG_URL')

micro_config_failover_url = os.getenv('MICRO_TEST_CONFIG_FAILOVER_URL')
if micro_config_failover_url is None:
    raise Exception('Cannot get environment variable MICRO_TEST_CONFIG_FAILOVER_URL')

MICRO_TEST_CONFIG = getjson(url=micro_config_url, failover_url=micro_config_failover_url)



from microprediction.univariate.distmachine import DistMachine
from tdigest import TDigest


class DigestDist(DistMachine):

    # Default DistMachine used in the SequentialStreamCrawler
    # This has no params

    def __init__(self, state: TDigest = None, params=None):
        state = state or TDigest()
        super().__init__(state=state, params=params)

    def update(self, value=None, dt=None, **kwargs):
        if value is not None:
            self.state.update(value)

    def inv_cdf(self, p):
        return self.state.percentile(100. * p)



from microconventions import MicroConventions, api_url
import requests, time, sys
from pprint import pprint
import numpy as np
import json
from typing import List, Tuple

# New video tutorials are available at https://www.microprediction.com/python-1 to help you
# get started reading historical data, and the like.


class MicroReader(MicroConventions):

    def __init__(self, base_url=None, **kwargs):
        """ Establish connection and adopt configuration parameters from site """
        super().__init__(base_url=base_url or api_url(), **kwargs)

    def fix_stream_name(self,name:str)->str:
        if len(name)<5 or name[-5:]!='.json':
            print('Stream names should end in .json')
            return name+'.json'
        else:
            return name

    def request_get_json(self, method, arg=None, data=None, throw=True):
        # TODO: Can remove this after microconventions>0.1.0
        try:
            if data is not None:
                res = requests.get(self.base_url + '/' + method + '/' + arg, data=data)
            elif arg is not None:
                res = requests.get(self.base_url + '/' + method + '/' + arg)
            elif data is None and arg is None:
                res = requests.get(self.base_url + '/' + method)
            if res.status_code == 200:
                return res.json()
        except ConnectionError as e:
            print('WARNING: ConnectionError attempting to get ' + method)
            if throw:
                raise e

    def __repr__(self):
        return json.dumps({'base_url':self.base_url})

    def get(self, name, throw=True):
        return self.request_get_json(method='live', arg=name, throw=throw)

    def get_current_value(self, name, throw=True):
        name = self.fix_stream_name(name=name)
        return self.request_get_json(method='live', arg=name, throw=throw)

    def get_sponsors(self) -> dict:
        return self.get_streams_by_sponsor()

    def get_streams_by_sponsor(self):
        return self.request_get_json(method='sponsors')

    def get_budgets(self):
        return self.get_streams_by_budget()

    def get_streams_by_budget(self):
        return self.request_get_json(method='budgets')

    def get_streams(self) -> dict:
        return self.get_sponsors()

    def get_prizes(self) -> dict:
        return self.request_get_json(method='prizes')

    def get_stream_names(self) -> [str]:
        return [ name for name in self.get_streams() ]

    def get_summary(self, name):
        return self.request_get_json(method='live', arg='summary::' + name)
        # res = requests.get(self.base_url + '/live/summary::' + name)

    def get_lagged(self,name, count=1000):
        name = self.fix_stream_name(name=name)
        return self.request_get_json(method='lagged',arg=name, data={'count':count-1})

    def get_recent_lagged_values(self, name, seconds:int, count=2000 ):
        lagged = self.get_lagged(name=name, count=count)
        t_cutoff = time.time() - seconds
        lagged_values = [v for (t, v) in lagged if t > t_cutoff]
        return lagged_values

    def append_chrono(self, name, chrono:List[Tuple], seconds=24 * 60 * 60, count=2000):
        """ Splice existing collection of historical time-value pairs to recent history stored on server

            chrono:  [(t,v)] history of values, preferably in chronological ordering
            returns: [(t,v)] extended history to present
        """
        chrono = sorted( chrono )
        lagged = self.get_recent_lagged_values(name=name, seconds=seconds, count=count)
        if len(chrono):
            t_cutoff = max([ t for (t,v) in chrono])
        else:
            t_cutoff = 0
        lagged_new = sorted( [ (t,v) for (t,v) in lagged if t>1e-6+t_cutoff ] )
        return list(chrono)+list(lagged_new)

    def get_lagged_values_and_times(self, name, count=1000):
        """ Preferred method """
        name = self.fix_stream_name(name=name)
        lagged = self.get_lagged(name=name, count=count)
        lagged_values = [l[1] for l in lagged]
        lagged_times  = [l[0] for l in lagged]
        return lagged_values, lagged_times

    def get_lagged_values(self, name:str, count:int=1000):
        """ Retrieve lagged values of a time series
        :param name:    cop.json   z1~cop.json   z2~cop~qp.json
        :return: [ float ]
        """
        name = self.fix_stream_name(name=name)
        lagged_values, lagged_times = self.get_lagged_values_and_times(name=name,count=count)
        return lagged_values

    def get_lagged_copulas(self, name:str, count:int=5000):
        """ Retrieve history of implied copulas in [0,1]^n
             returns [ [p1,p2,p3] ]
        """
        name = self.fix_stream_name(name=name)
        assert '~' in name,'This method is intended for copula streams'
        lagged_values, lagged_times = self.get_lagged_values_and_times(name=name, count=count)
        dim = 2 if 'z2~' in name else 3
        lagged_prctls = [ self.from_zcurve(zvalue,dim=dim) for zvalue in lagged_values ]
        return lagged_prctls

    def get_lagged_zvalues(self, name:str, count:int=5000):
        """ Retrieve history of implied z in [-inf,inf]^n
             returns [ [z1,z2,z3], [ , ,] ]
        """
        name = self.fix_stream_name(name=name)
        assert '~' in name, 'This method is intended for bivariate or trivariate copula streams'
        lagged_values, lagged_times = self.get_lagged_values_and_times(name=name, count=count)
        dim = 2 if 'z2~' in name else 3

        def expand(z):
            ps = self.from_zcurve(zvalue=z,dim=dim)
            return [ self.norminv(p) for p in ps ]

        lagged_zs = [ expand(z) for z in lagged_values]
        return lagged_zs


    def get_lagged_times(self, name:str, count:int=1000) -> list:
        """ Retrieve lagged times
        :param name:    cop.json   z1~cop.json   z2~cop~qp.json
        :return: [ float ]
        """
        name = self.fix_stream_name(name=name)
        lagged_values, lagged_times = self.get_lagged_values_and_times(name=name, count=count)
        return lagged_times

    def get_delayed_value(self, name: str, delay: int):
        """ Retrieve quarantined value.
            This is the most recent data point after going back at least delay seconds

            param: name    cop.json   z1~cop.json   z2~cop~qp.json
            param: delay
            :return: [ float ]
        """
        name = self.fix_stream_name(name=name)
        return self.request_get_json(method='live', arg='delayed::' + str(delay) + self.SEP + name)
        # res = requests.get(self.base_url + '/live/delayed::' + str(delay) + "::" + name)

    def get_repository(self, write_key):
        """ Get repository associated with a write key """
        # You can also supply a hash of the write key instead
        return self.request_get_json(method='repository', arg=write_key)
        # res = requests.get(self.base_url + '/repository/' + write_key)

    def get_predictions(self, write_key, name, delay:int, strip=True, consolidate=True):
        """ Retrieve predictions for a given horizon

              strip_percentiles  If false, returns dictionary with individual submissions
              consolidate        If false, returns tuples (owner,value)
                                 Otherwise just returns values

        """
        name = self.fix_stream_name(name=name)
        tickets = self.request_get_json(method='predictions', arg=name, data={"write_key":write_key,"delay":delay})
        if strip:
            tups = [(ticket.split('::')[1], val) for ticket, val in tickets.items()]
            if consolidate:
                return sorted([v for owner,v in tups])
            else:
                return tups
        else:
            return tickets




    def median(self, name: str, delay: int):
        name = self.fix_stream_name(name=name)
        return self.inv_cdf(name=name, delay=delay, p=0.5, num=15)

    def inv_cdf(self, name: str, delay: int, p=None, ps=None, num=25):
        """ Approximate PPF

             p    float  or
             ps  [float]
             num Number of interpolation points to use

        """
        name = self.fix_stream_name(name=name)
        # This won't choose x values based on the percentiles supplied, so it is a bit dumb in that sense
        cdf = self.get_cdf_lagged(name=name, delay=delay, num=num)
        ps_ = ps or [p]
        xs = np.interp(x=ps_, xp=cdf["x"], fp=cdf["y"], left=None, right=None)
        return list(xs) if ps is None else ps[0]

    def get_discrete_pdf_lagged(self, name: str, delay: int = None, num: int=25, lagged_values=None):
        """ Retrieve estimate of PDF ... only when discrete values are taken

                num   Maximum number of points to compute PDF at

        """
        if delay is None:
            delay = self.DELAYS[0]
        # PDF for continuous case is not implemented yet, sorry!
        lagged_values = lagged_values or self.get_lagged_values(name=name)
        values = self.cdf_values(lagged_values=lagged_values, num=num, as_discrete=True)
        raw_cdf = self._get_cdf(name=name, delay=delay, values=values)
        return {'x':raw_cdf['x'], 'y':self.discrete_pdf(raw_cdf['y'])} if raw_cdf.get('x') else raw_cdf

    def get_cdf_lagged(self, name: str, delay: int, num: int = 25, lagged_values=None, as_discrete=None):
        """ Get CDF using automatically selected x-values based on lags

              num:             Maximum number of points to use
              lagged_values:   Supply these to avoid an extra http round trip
              as_discrete      Supply bool if you know

        """
        lagged_values = lagged_values or self.get_lagged_values(name=name)
        if as_discrete is None:
            as_discrete = self.is_discrete(lagged_values=lagged_values, num=num, ndigits=12)
        values = self.cdf_values(lagged_values=lagged_values, num=num, as_discrete=as_discrete)
        raw_cdf = self._get_cdf(name=name, delay=delay, values=values)
        return self.discrete_cdf(raw_cdf) if as_discrete and raw_cdf.get('x') else raw_cdf

    def get_cdf(self, name: str, delay: int, values: [float], as_discrete=False) -> dict:
        """
            Get CDF using supplied x values
        """
        raw_cdf = self._get_cdf(name=name, delay=delay, values=values)
        return self.discrete_cdf(raw_cdf) if as_discrete and raw_cdf.get('x') else raw_cdf

    def _get_cdf(self, name: str, delay: int, values: [float]) -> dict:
        """ Implements approximate cumulative distribution function based on community micropredictions

              values  A list of x-values at which CDF will be evaluated
              Returns dict   {'x':[ float ], 'y':[ float ]}  where

            This cdf can be tricky to interpret
        """
        comma_sep_values = ",".join([str(v) for v in sorted(values)])
        return self.request_get_json(method='cdf', arg=name, data={'delay': delay, 'values': comma_sep_values})
        # res = requests.get(self.base_url + '/cdf/' + name, params={"delay":delay,"values": comma_sep_values})

    # For convenience...
    # This will move to microconventions

    def percentiles(self) -> [float]:
        """ A list of 225 evenly spaced numbers in (0,1) """
        return self.evenly_spaced_percentiles(num=self.num_predictions)


class MicroReaderStatus(MicroReader):

    def __init__(self):
        super().__init__()

    def reader_status(self):
        examples = {'get': {'name': 'cop.json'},
                    'get_current_value': {'name': 'cop.json'},
                    'get_sponsors': {},
                    'get_streams': {},
                    'get_budgets': {},
                    'get_summary': {'name': 'cop.json'},
                    'get_lagged_values': {'name': 'cop.json'},
                    'get_lagged_times': {'name': 'cop.json'},
                    'get_delayed_value': {'name': 'cop.json','delay':self.DELAYS[2]},
                    'get_cdf': {'name': 'cop.json','delay':self.DELAYS[2],'values':[-1,0,1]}
                    }
        report = list()
        for method, kwargs in examples.items():
            call_time = time.time()
            try:
                data = self.__getattribute__(method)(**kwargs)
                sz = sys.getsizeof(data)
                st = 'up' if sz > 0 else ''
                tm = time.time() - call_time
                er = ''
            except Exception as e:
                st = 'down'
                tm = -1
                er = str(e)
                sz = -1
            report.append((method, st, tm, er, sz))
        return report


if __name__ == "__main__":
    pprint(MicroReaderStatus().reader_status())



# New video tutorials are available at https://www.microprediction.com/python-1 to help you
# get started creating streams (see the 4th module in particular)

import logging
import requests
import time
import pytz
import random
from datetime import datetime, timedelta
from pprint import  pprint
from microprediction import MicroWriter
from microprediction.config_private import TRAFFIC_WRITE_KEY


logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


# Number of bikes available near large Manhattan hospitals
# longitude and latitude of rectangles within 500 feet of hospital
# citibike map: https://member.citibikenyc.com/map/
COORDS = [
    ((40.787880, -73.955862), (40.792001, -73.948389)),     # Mount Sinai - station_ids: 3345, 3299, 3363
    ((40.763573, -73.959175), (40.766666, -73.950179)),     # New York Presbytarian Upper East Side - station_ids: 3156, 3141
    ((40.735625, -73.979437), (40.743511, -73.970913)),     # Bellevue and NYU Langone - station_ids: 545, 3687, 174, 528, 536, 2012
    ((40.709229, -74.008624), (40.712033, -74.002787)),     # New York Presbytarian Lower Manhattan - station_ids: 224, 316
    ((40.783116, -73.949118), (40.787085, -73.939771)),     # Metropolitan Hospital Center - station_ids: 3309, 3318, 3338, 3351
]


def wait_between_attempts():
    """ Incremental backoff between connection attempts """
    wait_time = 19.3  # time is in seconds
    while True:
        yield wait_time
        wait_time = min(wait_time * 1.5, 30)
        wait_time *= (100 + random.randint(0, 50)) / 100

wait_time = wait_between_attempts()


def get_station_ids():
    """ Given COORDS, return all station_ids that fall within the rectangles """
    station_ids = []
    while True:
        try:
            r = requests.get("https://feeds.citibikenyc.com/stations/stations.json")
            if r.status_code==200:
                data = r.json()
                for c1, c2 in COORDS:
                    station_data = [ 
                        d for d in data["stationBeanList"] \
                        if d["latitude"] > c1[0] and d["latitude"] < c2[0] \
                        and d["longitude"] > c1[1] and d["longitude"] < c2[1]
                    ]
                    for station in station_data:
                        station_ids.append(int(station["id"]))
                return station_ids
        except requests.exceptions.RequestException as e:
            logger.error("Connection error %s: reconnecting..." % e)
            time.sleep(next(wait_time))
    return station_ids

def fetch_live_data(keys,field):
    """ Given list of station_ids as input, returns list of station[field] """
    while True:
        try:    
            r = requests.get("https://feeds.citibikenyc.com/stations/stations.json")
            if r.status_code==200:
                data = r.json()
                station_data = [ d for d in data["stationBeanList"] if int(d["id"]) in keys ]
                return [ station[field] for station in station_data ]
        except requests.exceptions.RequestException as e:
            logger.error("Connection error %s: reconnecting..." % e)
            time.sleep(next(wait_time))


def hospital_bike_activity(station_ids):
    prev = fetch_live_data(keys=station_ids,field="availableBikes")
    end_time = datetime.now() + timedelta(minutes = 20)
    activity = 0
    # measure activity change every 2 minutes for 20 minutes
    while datetime.now() < end_time:
        time.sleep(2 * 60)
        curr = fetch_live_data(keys=station_ids,field="availableBikes")
        for i in range(len(station_ids)):
            activity += abs(curr[i] - prev[i])
        prev = curr
    return activity



NAME = 'hospital_bike_activity.json'

station_ids = get_station_ids()
print("Initial activity is " + str(0), flush=True)


def run():
    try:
        mw = MicroWriter(write_key=TRAFFIC_WRITE_KEY)
    except:
        raise Exception("You need to set the write key for this example to work")
    while True:
        value = hospital_bike_activity(station_ids)
        utc_now = pytz.utc.localize(datetime.utcnow())
        pst_now = utc_now.astimezone(pytz.timezone("America/Los_Angeles"))
        res = mw.set(name=NAME,value=float(value))
        pprint({'PST time':pst_now.strftime("%H:%M"),'activity':value,"res":res})
        print('',flush=True)


if __name__=="__main__":
    run()



import pandas as pd

# New video tutorials are available at https://www.microprediction.com/python-1 to help you
# get started creating streams (see the 4th module in particular)


def height():
    df = pd.read_csv('https://www.ndbc.noaa.gov/data/realtime2/21413.dart')
    return float(df.iloc[1,:].values[0].split(' ')[-1])


if __name__=='__main__':
    print("Water height is "+str(height()))